@@ -21,6 +21,7 @@
 
 #include "csync_owncloud.h"
 
+#include <inttypes.h>
 
 /*
  * free the fetchCtx
@@ -59,7 +60,6 @@ int _connected = 0;                   /* flag to indicate if a connection exists
                                      the dav_session is valid */
 
 csync_auth_callback _authcb;
-csync_progress_callback _progresscb;
 long long chunked_total_size = 0;
 long long chunked_done = 0;
 
@@ -93,6 +93,11 @@ static void clean_caches() {
 char _buffer[PUT_BUFFER_SIZE];
 
 
+/* Differance in usec between two time  */
+static int64_t _timediff(struct timeval x , struct timeval y)
+{
+    return (y.tv_sec - x.tv_sec)*1000000 + (y.tv_usec - x.tv_usec);
+}
 
 /*
  * helper method to build up a user text for SSL problems, called from the
@@ -124,7 +129,6 @@ static int verify_sslcert(void *userdata, int failures,
     int ret = -1;
     const ne_ssl_certificate *cert = certificate;
 
-    (void) userdata;
     (void) userdata;
     memset( problem, 0, LEN );
 
@@ -165,7 +169,7 @@ static int verify_sslcert(void *userdata, int failures,
         /* call the csync callback */
         DEBUG_WEBDAV("Call the csync callback for SSL problems");
         memset( buf, 0, NE_ABUFSIZ );
-        (*_authcb) ( problem, buf, NE_ABUFSIZ-1, 1, 0, dav_session.userdata );
+        (*_authcb) ( problem, buf, NE_ABUFSIZ-1, 1, 0, NULL );
         if( buf[0] == 'y' || buf[0] == 'Y') {
             ret = 0;
         } else {
@@ -204,12 +208,12 @@ static int ne_auth( void *userdata, const char *realm, int attempt,
             /* call the csync callback */
             DEBUG_WEBDAV("Call the csync callback for %s", realm );
             memset( buf, 0, NE_ABUFSIZ );
-            (*_authcb) ("Enter your username: ", buf, NE_ABUFSIZ-1, 1, 0, dav_session.userdata );
+            (*_authcb) ("Enter your username: ", buf, NE_ABUFSIZ-1, 1, 0, NULL );
             if( strlen(buf) < NE_ABUFSIZ ) {
                 strcpy( username, buf );
             }
             memset( buf, 0, NE_ABUFSIZ );
-            (*_authcb) ("Enter your password: ", buf, NE_ABUFSIZ-1, 0, 0, dav_session.userdata );
+            (*_authcb) ("Enter your password: ", buf, NE_ABUFSIZ-1, 0, 0, NULL );
             if( strlen(buf) < NE_ABUFSIZ) {
                 strcpy( password, buf );
             }
@@ -370,7 +374,7 @@ static void post_request_hook(ne_request *req, void *userdata, const ne_status *
 
 /*
  * this hook is called just after a request has been created, before its sent.
- * Here it is used to set the session cookie if available.
+ * Here it is used to set the proxy connection header if available.
  */
 static void request_created_hook(ne_request *req, void *userdata,
                                  const char *method, const char *requri)
@@ -381,32 +385,95 @@ static void request_created_hook(ne_request *req, void *userdata,
 
     if( !req ) return;
 
-    if(dav_session.session_key) {
-        /* DEBUG_WEBDAV("Setting PHPSESSID to %s", dav_session.session_key); */
-        ne_add_request_header(req, "Cookie", dav_session.session_key);
-    }
-
     if(dav_session.proxy_type) {
         /* required for NTLM */
         ne_add_request_header(req, "Proxy-Connection", "Keep-Alive");
     }
 }
 
+/*
+ * this hook is called just before a request has been sent.
+ * Here it is used to set the session cookie if available.
+ */
+static void pre_send_hook(ne_request *req, void *userdata,
+                          ne_buffer *header)
+{
+    (void) userdata;
+
+    if( !req ) return;
+
+    if(dav_session.session_key) {
+        ne_buffer_concat(header, "Cookie: ", dav_session.session_key, "\r\n", NULL);
+    }
+}
+
+static int post_send_hook(ne_request *req, void *userdata,
+                          const ne_status *status)
+{
+    const char *location;
+
+    (void) userdata;
+    (void) status;
+
+    location = ne_get_response_header(req, "Location");
+
+    if( !location ) return NE_OK;
+
+    if( dav_session.redir_callback ) {
+        if( dav_session.redir_callback( dav_session.csync_ctx, location ) ) {
+            return NE_REDIRECT;
+        } else {
+            return NE_RETRY;
+        }
+    }
+
+    return NE_REDIRECT;
+}
+
 /* called from neon */
 static void ne_notify_status_cb (void *userdata, ne_session_status status,
                                  const ne_session_status_info *info)
 {
     struct transfer_context *tc = (struct transfer_context*) userdata;
+    struct timeval now;
+    int bandwidth_limit = 0;
 
-    if (_progresscb && (status == ne_status_sending || status == ne_status_recving)) {
-        if (info->sr.total > 0)
-            _progresscb(tc->url, CSYNC_NOTIFY_PROGRESS,
+    if ((status == ne_status_sending || status == ne_status_recving)) {
+        if (info->sr.total > 0) {
+          oc_notify_progress( tc->url, CSYNC_NOTIFY_PROGRESS,
                         chunked_done + info->sr.progress,
-                        chunked_total_size ? chunked_total_size : info->sr.total,
-                        dav_session.userdata);
-
-        if (chunked_total_size && info->sr.total == info->sr.progress)
+                        chunked_total_size ? chunked_total_size : info->sr.total );
+        }
+        if (chunked_total_size && info->sr.total > 0 && info->sr.total == info->sr.progress) {
             chunked_done += info->sr.total;
+            DEBUG_WEBDAV("Chunk upload completed for '%s' (%lld bytes out of %lld)", tc->url , chunked_done, chunked_total_size);
+        }
+    }
+
+    /* throttle connection */
+    if (status == ne_status_sending) bandwidth_limit = dav_session.bandwidth_limit_upload;
+    if (status == ne_status_recving) bandwidth_limit = dav_session.bandwidth_limit_download;
+    if (bandwidth_limit > 0 && gettimeofday(&now, NULL) == 0) {
+        int64_t diff = _timediff(tc->last_time, now);
+        int64_t len = info->sr.progress - tc->last_progress;
+        if (len > 0 && diff > 0 && (1000000 * len / diff) > (int64_t)bandwidth_limit) {
+            int64_t wait_time = (1000000 * len / bandwidth_limit) - diff;
+            if (wait_time > 0) {
+                usleep(wait_time);
+            }
+        }
+        tc->last_progress = info->sr.progress;
+        gettimeofday(&tc->last_time, NULL);
+    } else if (bandwidth_limit < 0 && bandwidth_limit > -100 && gettimeofday(&now, NULL) == 0) {
+        int64_t diff = _timediff(tc->last_time, now);
+        if (diff > 0) {
+            // -bandwidth_limit is the % of bandwidth
+            int64_t wait_time = -diff * (1 + 100.0 / bandwidth_limit);
+            if (wait_time > 0) {
+                usleep(wait_time);
+            }
+        }
+        gettimeofday(&tc->last_time, NULL);
     }
 }
 
@@ -518,12 +585,15 @@ static int dav_connect(const char *base_url) {
         ne_ssl_trust_default_ca( dav_session.ctx );
         ne_ssl_set_verify( dav_session.ctx, verify_sslcert, 0 );
     }
-    ne_redirect_register( dav_session.ctx );
 
-    /* Hook to get the Session ID */
-    ne_hook_post_headers( dav_session.ctx, post_request_hook, NULL );
-    /* Hook called when a request is built. It sets the PHPSESSID header */
+    /* Hook called when a request is created. It sets the proxy connection header. */
     ne_hook_create_request( dav_session.ctx, request_created_hook, NULL );
+    /* Hook called after response headers are read. It gets the Session ID. */
+    ne_hook_post_headers( dav_session.ctx, post_request_hook, NULL );
+    /* Hook called before a request is sent. It sets the cookies. */
+    ne_hook_pre_send( dav_session.ctx, pre_send_hook, NULL );
+    /* Hook called after request is dispatched. Used for handling possible redirections. */
+    ne_hook_post_send( dav_session.ctx, post_send_hook, NULL );
 
     /* Proxy support */
     proxystate = configureProxy( dav_session.ctx );
@@ -691,9 +761,9 @@ static struct listdir_context *fetch_resource_list(const char *uri, int depth)
                          req_status->reason_phrase);
             ret = NE_CONNECT;
             set_error_message(req_status->reason_phrase);
-            if (_progresscb) {
-                _progresscb(uri, CSYNC_NOTIFY_ERROR,  req_status->code, (long long)(req_status->reason_phrase) ,dav_session.userdata);
-            }
+            oc_notify_progress( uri, CSYNC_NOTIFY_ERROR,
+                                req_status->code,
+                                (intptr_t)(req_status->reason_phrase) );
         }
         DEBUG_WEBDAV("Simple propfind result code %d.", req_status->code);
     } else {
@@ -730,16 +800,6 @@ static struct listdir_context *fetch_resource_list(const char *uri, int depth)
     if( hdl )
         ne_propfind_destroy(hdl);
 
-    if( ret == NE_REDIRECT ) {
-        const ne_uri *redir_ne_uri = NULL;
-        char *redir_uri = NULL;
-        redir_ne_uri = ne_redirect_location(dav_session.ctx);
-        if( redir_ne_uri ) {
-            redir_uri = ne_uri_unparse(redir_ne_uri);
-            DEBUG_WEBDAV("Permanently moved to %s", redir_uri);
-        }
-    }
-
     if( ret != NE_OK ) {
         free_fetchCtx(fetchCtx);
         return NULL;
@@ -911,8 +971,9 @@ static int content_reader(void *userdata, const char *buf, size_t len)
        /* DEBUG_WEBDAV("Writing %scompressed %d bytes", (writeCtx->decompress ? "" : "NON "), len); */
        written = write(writeCtx->fd, buf, len);
        if( len != written ) {
-           DEBUG_WEBDAV("WRN: content_reader wrote wrong num of bytes: %zu, %zu", len, written);
+           DEBUG_WEBDAV("WRN: content_reader wrote wrong num of bytes: %zu, %zu %d %d", len, written, errno, writeCtx->fd);
        }
+       writeCtx->get_size += len;
        return NE_OK;
    } else {
      errno = EBADF;
@@ -1038,6 +1099,13 @@ static const char* owncloud_file_id( const char *path )
             cbuf = c_strdup(header);
         }
     }
+
+    /* fix server problem: If we end up with an empty string, set something strange... */
+    if( c_streq(cbuf, "") || c_streq(cbuf, "\"\"") ) {
+        SAFE_FREE(cbuf);
+        cbuf = c_strdup("empty_etag");
+    }
+
     DEBUG_WEBDAV("Get file ID for %s: %s", path, cbuf ? cbuf:"<null>");
     if( fs ) csync_vio_file_stat_destroy(fs);
     if( req ) ne_request_destroy(req);
@@ -1152,6 +1220,11 @@ static int _user_want_abort()
     return csync_abort_requested(dav_session.csync_ctx);
 }
 
+static void _log_callback(const char *func, const char *text)
+{
+    csync_log(dav_session.csync_ctx, CSYNC_LOG_PRIORITY_TRACE, func, "%s", text);
+}
+
 static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_handle_t *hdl ) {
     int rc  = 0;
     int neon_stat;
@@ -1160,8 +1233,8 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
     fhandle_t *fh = (fhandle_t *) src;
     int fd;
     int error_code = 0;
-    const char *error_string = NULL;
     char *clean_uri = NULL;
+    int64_t file_size;
 
     if( ! write_ctx ) {
         errno = EINVAL;
@@ -1179,7 +1252,9 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
     chunked_total_size = 0;
     chunked_done = 0;
 
-    DEBUG_WEBDAV("Sendfile handling request type %s.", write_ctx->method);
+    gettimeofday(&write_ctx->last_time, NULL);
+
+    DEBUG_WEBDAV("Sendfile handling request type %s. fd %d", write_ctx->method, fd);
 
     /*
      * Copy from the file descriptor if method == PUT
@@ -1198,6 +1273,13 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
       do {
         Hbf_State state = HBF_SUCCESS;
         hbf_transfer_t *trans = hbf_init_transfer(clean_uri);
+        hbf_set_log_callback(trans, _log_callback);
+        if (dav_session.hbf_block_size > 0) {
+          trans->threshold = trans->block_size = dav_session.hbf_block_size;
+        }
+        if (dav_session.hbf_threshold > 0) {
+          trans->threshold = dav_session.hbf_threshold;
+        }
         finished = true;
 
         if (!trans) {
@@ -1205,6 +1287,10 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
           rc = 1;
         } else {
           state = hbf_splitlist(trans, fd);
+          file_size = trans->stat_size;
+          (void) file_size;
+          DEBUG_WEBDAV("about to send %d block", trans->block_cnt);
+
 
           /* Reuse chunk info that was stored in database if existing. */
           if (dav_session.chunk_info && dav_session.chunk_info->transfer_id) {
@@ -1213,9 +1299,9 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
             trans->transfer_id = dav_session.chunk_info->transfer_id;
           }
 
-          if (state == HBF_SUCCESS && _progresscb) {
+          if (state == HBF_SUCCESS) {
             ne_set_notifier(dav_session.ctx, ne_notify_status_cb, write_ctx);
-            _progresscb(write_ctx->url, CSYNC_NOTIFY_START_UPLOAD, 0 , 0, dav_session.userdata);
+            oc_notify_progress(write_ctx->url, CSYNC_NOTIFY_START_UPLOAD, 0, 0);
           }
 
           /* Register the abort callback */
@@ -1245,8 +1331,11 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
             }
 
             if( finished ) {
-              error_string = hbf_error_string(state);
+              SAFE_FREE(dav_session.error_string);
+              dav_session.error_string = c_strdup(hbf_error_string(trans, state));
               error_code = hbf_fail_http_code(trans);
+              set_errno_from_http_errcode(error_code);
+
               rc = 1;
               if (dav_session.chunk_info) {
                 dav_session.chunk_info->start_id = trans->start_id;
@@ -1258,11 +1347,11 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
         hbf_free_transfer(trans);
       } while( !finished );
 
-      if (_progresscb) {
-        ne_set_notifier(dav_session.ctx, 0, 0);
-        _progresscb(write_ctx->url, rc != 0 ? CSYNC_NOTIFY_ERROR :
-                                              CSYNC_NOTIFY_FINISHED_UPLOAD, error_code,
-                    (long long)(error_string), dav_session.userdata);
+      ne_set_notifier(dav_session.ctx, 0, 0);
+      if( rc == 0 ) {
+        DEBUG_WEBDAV(" * Upload finished %s", write_ctx->url);
+      } else {
+        oc_notify_progress(write_ctx->url, CSYNC_NOTIFY_ERROR, error_code, (intptr_t)(dav_session.error_string));
       }
     } else if( c_streq( write_ctx->method, "GET") ) {
       /* GET a file to the file descriptor */
@@ -1270,10 +1359,8 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
       int retry = 0;
       DEBUG_WEBDAV("  -- GET on %s", write_ctx->url);
       write_ctx->fd = fd;
-      if (_progresscb) {
-        ne_set_notifier(dav_session.ctx, ne_notify_status_cb, write_ctx);
-        _progresscb(write_ctx->url, CSYNC_NOTIFY_START_DOWNLOAD, 0 , 0, dav_session.userdata);
-      }
+      ne_set_notifier(dav_session.ctx, ne_notify_status_cb, write_ctx);
+      oc_notify_progress(write_ctx->url, CSYNC_NOTIFY_START_DOWNLOAD, 0 , 0);
 
       do {
         csync_stat_t sb;
@@ -1291,12 +1378,12 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
         /* Allow compressed content by setting the header */
         ne_add_request_header( write_ctx->req, "Accept-Encoding", "gzip" );
 
-        if (fstat(fd, &sb) >= 0 && sb.st_size > 0) {
+        if (_tfstat(fd, &sb) >= 0 && sb.st_size > 0) {
             char brange[64];
-            ne_snprintf(brange, sizeof brange, "bytes=%lld-", (long long) sb.st_size);
+            snprintf(brange, sizeof brange, "bytes=%"PRId64"-", (int64_t) sb.st_size);
             ne_add_request_header(write_ctx->req, "Range", brange);
             ne_add_request_header(write_ctx->req, "Accept-Ranges", "bytes");
-            DEBUG_WEBDAV("Retry with range %s", brange);
+            DEBUG_WEBDAV("Retry with range %s fd %d", brange, fd);
         }
 
         /* hook called before the content is parsed to set the correct reader,
@@ -1305,40 +1392,6 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
         ne_hook_post_headers( dav_session.ctx, install_content_reader, write_ctx );
 
         neon_stat = ne_request_dispatch(write_ctx->req );
-        /* possible return codes are:
-         *  NE_OK, NE_AUTH, NE_CONNECT, NE_TIMEOUT, NE_ERROR (from ne_request.h)
-         */
-
-        if( neon_stat != NE_OK ) {
-            if (neon_stat == NE_TIMEOUT && (++retry) < 3)
-                continue;
-
-            set_errno_from_neon_errcode(neon_stat);
-            DEBUG_WEBDAV("Error GET: Neon: %d, errno %d", neon_stat, errno);
-            error_string = dav_session.error_string;
-            error_code = errno;
-            rc = 1;
-        } else {
-            status = ne_get_status( write_ctx->req );
-            DEBUG_WEBDAV("GET http result %d (%s)", status->code, status->reason_phrase ? status->reason_phrase : "<empty");
-            if( status->klass != 2 ) {
-                DEBUG_WEBDAV("sendfile request failed with http status %d!", status->code);
-                set_errno_from_http_errcode( status->code );
-                /* decide if soft error or hard error that stops the whole sync. */
-                /* Currently all problems concerning one file are soft errors */
-                if( status->klass == 4 /* Forbidden and stuff, soft error */ ) {
-                    rc = 1;
-                } else if( status->klass == 5 /* Server errors and such */ ) {
-                    rc = 1; /* No Abort on individual file errors. */
-                } else {
-                    rc = 1;
-                }
-                error_code = status->code;
-                error_string = status->reason_phrase;
-            } else {
-                DEBUG_WEBDAV("http request all cool, result code %d", status->code);
-            }
-        }
 
         /* delete the hook again, otherwise they get chained as they are with the session */
         ne_unhook_post_headers( dav_session.ctx, install_content_reader, write_ctx );
@@ -1347,16 +1400,67 @@ static int owncloud_sendfile(csync_vio_method_handle_t *src, csync_vio_method_ha
         if( write_ctx->decompress ) {
             ne_decompress_destroy( write_ctx->decompress );
         }
+        /* possible return codes are:
+         *  NE_OK, NE_AUTH, NE_CONNECT, NE_TIMEOUT, NE_ERROR (from ne_request.h)
+         */
+
+        status = ne_get_status( write_ctx->req );
+
+        if( neon_stat != NE_OK ) {
+            /* If a timeout happened try again for three times */
+            if (neon_stat == NE_TIMEOUT && (++retry) < 3) {
+                continue;
+            }
+
+            set_errno_from_neon_errcode(neon_stat);
+            DEBUG_WEBDAV("Error GET: Neon: %d, errno %d, string %s", neon_stat, errno, dav_session.error_string);
+            error_code = errno;
+            if( status != NULL && errno != ERRNO_ERROR_STRING ) {
+                SAFE_FREE(dav_session.error_string);
+                dav_session.error_string = c_strdup(status->reason_phrase);
+            }
+            rc = 1;
+        } else {
+            if( status != NULL ) {
+                DEBUG_WEBDAV("GET http result %d (%s)", status->code, status->reason_phrase ? status->reason_phrase : "<empty");
+                if( status->klass != 2 ) {
+                    DEBUG_WEBDAV("sendfile request failed with http status %d!", status->code);
+                    set_errno_from_http_errcode( status->code );
+                    /* decide if soft error or hard error that stops the whole sync. */
+                    /* Currently all problems concerning one file are soft errors */
+                    if( status->klass == 4 /* Forbidden and stuff, soft error */ ) {
+                        rc = 1;
+                    } else if( status->klass == 5 /* Server errors and such */ ) {
+                        rc = 1; /* No Abort on individual file errors. */
+                    } else {
+                        rc = 1;
+                    }
+                    error_code = status->code;
+                    SAFE_FREE(dav_session.error_string);
+                    dav_session.error_string = c_strdup(status->reason_phrase);
+                } else {
+                    DEBUG_WEBDAV("http request all cool, result code %d", status->code);
+                }
+            } else {
+                /* No status but still a problem */
+                DEBUG_WEBDAV("GET failed, but no neon status available.");
+                error_code = 400;
+                rc = 1;
+            }
+        }
+
         break;
       } while (1);
-      if (_progresscb) {
-          ne_set_notifier(dav_session.ctx, 0, 0);
-          _progresscb(write_ctx->url, (rc != NE_OK) ? CSYNC_NOTIFY_ERROR :
-                      CSYNC_NOTIFY_FINISHED_DOWNLOAD, error_code ,
-                      (long long)(error_string), dav_session.userdata);
+
+      ne_set_notifier(dav_session.ctx, 0, 0);
+      if( rc == 0 ) {
+        DEBUG_WEBDAV(" ** Finished download %s", write_ctx->url);
+      } else {
+        oc_notify_progress( write_ctx->url, CSYNC_NOTIFY_ERROR, error_code , (intptr_t)(dav_session.error_string));
       }
     } else  {
         DEBUG_WEBDAV("Unknown method!");
+        errno = ERRNO_GENERAL_ERROR;
         rc = -1;
     }
 
@@ -1405,7 +1509,7 @@ static ssize_t owncloud_read(csync_vio_method_handle_t *fhandle, void *buf, size
     return len;
 }
 
-static off_t owncloud_lseek(csync_vio_method_handle_t *fhandle, off_t offset, int whence) {
+static int64_t owncloud_lseek(csync_vio_method_handle_t *fhandle, int64_t offset, int whence) {
     (void) fhandle;
     (void) offset;
     (void) whence;
@@ -1452,14 +1556,12 @@ static csync_vio_file_stat_t *owncloud_readdir(csync_vio_method_handle_t *dhandl
 
     struct listdir_context *fetchCtx = dhandle;
 
-    if( fetchCtx->currResource ) {
-        // DEBUG_WEBDAV("readdir method called for %s", fetchCtx->currResource->uri);
-    } else {
+    if( fetchCtx == NULL) {
         /* DEBUG_WEBDAV("An empty dir or at end"); */
         return NULL;
     }
 
-    while( fetchCtx && fetchCtx->currResource ) {
+    while( fetchCtx->currResource ) {
         resource* currResource = fetchCtx->currResource;
         char *escaped_path = NULL;
 
@@ -1480,7 +1582,6 @@ static csync_vio_file_stat_t *owncloud_readdir(csync_vio_method_handle_t *dhandl
         }
 
         /* This is the target URI */
-        DEBUG_WEBDAV( "Skipping target resource.");
         SAFE_FREE( escaped_path );
     }
 
@@ -1519,9 +1620,9 @@ static int owncloud_mkdir(const char *uri, mode_t mode) {
          * To keep csync vio_mkdirs working errno EEXIST has to be returned. */
         if (errno == EPERM && http_result_code_from_session() == 405) {
             errno = EEXIST;
-        } else if (rc != NE_OK && _progresscb) {
-            _progresscb(uri, CSYNC_NOTIFY_ERROR,  http_result_code_from_session(),
-                        (long long)(dav_session.error_string) ,dav_session.userdata);
+        } else if (rc != NE_OK) {
+            oc_notify_progress(uri, CSYNC_NOTIFY_ERROR,  http_result_code_from_session(),
+                               (intptr_t)(dav_session.error_string) );
         }
     }
     SAFE_FREE( path );
@@ -1579,9 +1680,9 @@ static int owncloud_rename(const char *olduri, const char *newuri) {
             errno = ENOENT;
         } else {
             set_errno_from_neon_errcode(rc);
-            if (rc != NE_OK && _progresscb) {
-                _progresscb(olduri, CSYNC_NOTIFY_ERROR,  http_result_code_from_session(),
-                            (long long)(dav_session.error_string) ,dav_session.userdata);
+            if (rc != NE_OK) {
+                oc_notify_progress(olduri, CSYNC_NOTIFY_ERROR,  http_result_code_from_session(),
+                                   (intptr_t)(dav_session.error_string) );
             }
         }
     }
@@ -1723,10 +1824,6 @@ static int owncloud_set_property(const char *key, void *data) {
         dav_session.proxy_port = *(int*)(data);
         return 0;
     }
-    if (c_streq(key, "progress_callback")) {
-        _progresscb = *(csync_progress_callback*)(data);
-        return 0;
-    }
     if (c_streq(key, "read_timeout") || c_streq(key, "timeout")) {
         dav_session.read_timeout = *(int*)(data);
         return 0;
@@ -1748,6 +1845,34 @@ static int owncloud_set_property(const char *key, void *data) {
         dav_session.no_recursive_propfind = *(bool*)(data);
         return 0;
     }
+    if( c_streq(key, "hbf_block_size")) {
+        dav_session.hbf_block_size = *(int64_t*)(data);
+        return 0;
+    }
+    if( c_streq(key, "hbf_threshold")) {
+        dav_session.hbf_threshold = *(int64_t*)(data);
+        return 0;
+    }
+    if( c_streq(key, "bandwidth_limit_upload")) {
+        dav_session.bandwidth_limit_upload = *(int*)(data);
+        return 0;
+    }
+    if( c_streq(key, "bandwidth_limit_download")) {
+        dav_session.bandwidth_limit_download = *(int*)(data);
+        return 0;
+    }
+    if( c_streq(key, "overall_progress_data")) {
+      dav_session.overall_progress_data = (csync_overall_progress_t*)(data);
+    }
+    if( c_streq(key, "redirect_callback")) {
+        if (data) {
+            csync_owncloud_redirect_callback_t* cb_wrapper = data;
+
+            dav_session.redir_callback = *cb_wrapper;
+        } else {
+            dav_session.redir_callback = NULL;
+        }
+    }
 
     return -1;
 }
@@ -1784,12 +1909,12 @@ csync_vio_method_t *vio_module_init(const char *method_name, const char *args,
                                     csync_auth_callback cb, void *userdata) {
     (void) method_name;
     (void) args;
+    (void) userdata;
 
     _authcb = cb;
     _connected = 0;  /* triggers dav_connect to go through the whole neon setup */
 
     memset(&dav_session, 0, sizeof(dav_session));
-    dav_session.userdata = userdata;
 
     return &_method;
 }
