@@ -612,7 +612,7 @@ static ssize_t _sftp_write(csync_vio_method_handle_t *fhandle, const void *buf,
   return rc;
 }
 
-static off_t _sftp_lseek(csync_vio_method_handle_t *fhandle, off_t offset, int whence) {
+static int64_t _sftp_lseek(csync_vio_method_handle_t *fhandle, int64_t offset, int whence) {
   /* FIXME: really implement seek for lseek? */
   (void) whence;
   sftp_seek(fhandle, offset);
