@@ -56,6 +56,7 @@
 #include "vio/csync_vio.h"
 
 #include "csync_log.h"
+#include "csync_rename.h"
 
 static int _key_cmp(const void *key, const void *data) {
   uint64_t a;
@@ -153,6 +154,12 @@ int csync_create(CSYNC **csync, const char *local, const char *remote) {
     return -1;
   }
 
+  ctx->local.list     = 0;
+  ctx->remote.list    = 0;
+  ctx->current_fs = NULL;
+
+  ctx->abort = false;
+
   *csync = ctx;
   return 0;
 }
@@ -182,19 +189,17 @@ int csync_init(CSYNC *ctx) {
   }
 
   /* create lock file */
-  if (asprintf(&lock, "%s/%s", ctx->options.config_dir, CSYNC_LOCK_FILE) < 0) {
+  if (asprintf(&lock, "%s/%s", ctx->local.uri, CSYNC_LOCK_FILE) < 0) {
     ctx->error_code = CSYNC_ERR_MEM;
     rc = -1;
     goto out;
   }
 
-#ifndef _WIN32
   if (csync_lock(ctx, lock) < 0) {
     ctx->error_code = CSYNC_ERR_LOCK;
     rc = -1;
     goto out;
   }
-#endif
 
   /* load config file */
   if (asprintf(&config, "%s/%s", ctx->options.config_dir, CSYNC_CONF_FILE) < 0) {
@@ -216,7 +221,7 @@ int csync_init(CSYNC *ctx) {
   }
 
   if (csync_exclude_load(ctx, exclude) < 0) {
-    strerror_r(errno, errbuf, sizeof(errbuf));
+    C_STRERROR(errno, errbuf, sizeof(errbuf));
     CSYNC_LOG(CSYNC_LOG_PRIORITY_WARN, "Could not load %s - %s", exclude,
               errbuf);
   }
@@ -230,28 +235,11 @@ int csync_init(CSYNC *ctx) {
   }
 
   if (csync_exclude_load(ctx, exclude) < 0) {
-    strerror_r(errno, errbuf, sizeof(errbuf));
+    C_STRERROR(errno, errbuf, sizeof(errbuf));
     CSYNC_LOG(CSYNC_LOG_PRIORITY_INFO, "Could not load %s - %s", exclude, 
               errbuf);
   }
 
-
-  /* create/load statedb */
-  if (! csync_is_statedb_disabled(ctx)) {
-    rc = asprintf(&ctx->statedb.file, "%s/.csync_journal.db",
-                  ctx->local.uri);
-    if (rc < 0) {
-        goto out;
-    }
-    CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "Journal: %s", ctx->statedb.file);
-
-    if (csync_statedb_load(ctx, ctx->statedb.file) < 0) {
-      ctx->error_code = CSYNC_ERR_STATEDB_LOAD;
-      rc = -1;
-      goto out;
-    }
-  }
-
   ctx->local.type = LOCAL_REPLICA;
 
   /* check for uri */
@@ -329,12 +317,10 @@ retry_vio_init:
     }
   }
 
-  if (ctx->callbacks.progresscb)
-    csync_vio_set_property(ctx, "progress_callback", &ctx->callbacks.progresscb);
-
   if (ctx->options.timeout)
     csync_vio_set_property(ctx, "timeout", &ctx->options.timeout);
 
+  /* Install progress callbacks in the module. */
   if (c_rbtree_create(&ctx->local.tree, _key_cmp, _data_cmp) < 0) {
     ctx->error_code = CSYNC_ERR_TREE;
     rc = -1;
@@ -349,6 +335,10 @@ retry_vio_init:
 
   ctx->status = CSYNC_STATUS_INIT;
 
+  csync_lock_remove(ctx, lock);
+
+  csync_set_module_property(ctx, "csync_context", ctx);
+
   /* initialize random generator */
   srand(time(NULL));
 
@@ -364,13 +354,45 @@ out:
 int csync_update(CSYNC *ctx) {
   int rc = -1;
   struct timespec start, finish;
-  
+  char *lock = NULL;
+
   if (ctx == NULL) {
     errno = EBADF;
     return -1;
   }
   ctx->error_code = CSYNC_ERR_NONE;
 
+  /* try to create lock file */
+  if (asprintf(&lock, "%s/%s", ctx->local.uri, CSYNC_LOCK_FILE) < 0) {
+    ctx->error_code = CSYNC_ERR_MEM;
+    rc = -1;
+    return rc;
+  }
+
+  if (csync_lock(ctx, lock) < 0) {
+    ctx->error_code = CSYNC_ERR_LOCK;
+    rc = -1;
+    return rc;
+  }
+
+  SAFE_FREE(lock);
+
+  /* create/load statedb */
+  if (! csync_is_statedb_disabled(ctx)) {
+    rc = asprintf(&ctx->statedb.file, "%s/.csync_journal.db",
+                  ctx->local.uri);
+    if (rc < 0) {
+        return rc;
+    }
+    CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "Journal: %s", ctx->statedb.file);
+
+    if (csync_statedb_load(ctx, ctx->statedb.file) < 0) {
+      ctx->error_code = CSYNC_ERR_STATEDB_LOAD;
+      rc = -1;
+      return rc;
+    }
+  }
+
   csync_memstat_check(ctx);
 
   /* update detection for local replica */
@@ -379,6 +401,11 @@ int csync_update(CSYNC *ctx) {
   ctx->replica = ctx->local.type;
 
   rc = csync_ftw(ctx, ctx->local.uri, csync_walker, MAX_DEPTH);
+  if (rc < 0) {
+    if(ctx->error_code == CSYNC_ERR_NONE)
+        ctx->error_code = csync_errno_to_csync_error(CSYNC_ERR_UPDATE);
+    return -1;
+  }
 
   csync_gettime(&finish);
 
@@ -387,11 +414,6 @@ int csync_update(CSYNC *ctx) {
       c_secdiff(finish, start), c_rbtree_size(ctx->local.tree));
   csync_memstat_check(ctx);
 
-  if (rc < 0) {
-    if(ctx->error_code == CSYNC_ERR_NONE)
-        ctx->error_code = csync_errno_to_csync_error(CSYNC_ERR_UPDATE);
-    return -1;
-  }
 
   /* update detection for remote replica */
   if( ! ctx->options.local_only_mode ) {
@@ -400,6 +422,11 @@ int csync_update(CSYNC *ctx) {
       ctx->replica = ctx->remote.type;
 
       rc = csync_ftw(ctx, ctx->remote.uri, csync_walker, MAX_DEPTH);
+      if (rc < 0) {
+          if(ctx->error_code == CSYNC_ERR_NONE )
+            ctx->error_code = csync_errno_to_csync_error(CSYNC_ERR_UPDATE);
+          return -1;
+      }
       csync_gettime(&finish);
 
       CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG,
@@ -408,11 +435,6 @@ int csync_update(CSYNC *ctx) {
                 c_secdiff(finish, start), c_rbtree_size(ctx->remote.tree));
       csync_memstat_check(ctx);
 
-      if (rc < 0) {
-          if(ctx->error_code == CSYNC_ERR_NONE )
-            ctx->error_code = csync_errno_to_csync_error(CSYNC_ERR_UPDATE);
-          return -1;
-      }
   }
   ctx->status |= CSYNC_STATUS_UPDATE;
 
@@ -484,6 +506,22 @@ int csync_propagate(CSYNC *ctx) {
   }
   ctx->error_code = CSYNC_ERR_NONE;
 
+  rc = csync_init_progress(ctx);
+  if (rc < 0) {
+    if( ctx->error_code == CSYNC_ERR_NONE )
+        ctx->error_code = csync_errno_to_csync_error( CSYNC_ERR_PROPAGATE);
+    return -1;
+  }
+
+  ctx->current = REMOTE_REPLICA;
+  ctx->replica = ctx->remote.type;
+  rc = csync_propagate_renames(ctx);
+  if (rc < 0) {
+      if( ctx->error_code == CSYNC_ERR_NONE )
+          ctx->error_code = csync_errno_to_csync_error( CSYNC_ERR_PROPAGATE);
+      return -1;
+  }
+
   /* Reconciliation for local replica */
   csync_gettime(&start);
 
@@ -518,35 +556,14 @@ int csync_propagate(CSYNC *ctx) {
       "Propagation for remote replica took %.2f seconds visiting %zu files.",
       c_secdiff(finish, start), c_rbtree_size(ctx->remote.tree));
 
+  csync_finalize_progress(ctx);
+
   if (rc < 0) {
     if( ctx->error_code == CSYNC_ERR_NONE )
       ctx->error_code = csync_errno_to_csync_error( CSYNC_ERR_PROPAGATE);
     return -1;
   }
 
-  csync_gettime(&start);
-  ctx->current = LOCAL_REPLICA;
-  ctx->replica = ctx->local.type;
-  rc = csync_correct_id(ctx);
-  if (rc < 0) {
-      ctx->error_code = CSYNC_ERR_PROPAGATE;
-      return -1;
-  }
-
-  ctx->current = REMOTE_REPLICA;
-  ctx->replica = ctx->remote.type;
-  rc = csync_correct_id(ctx);
-  if (rc < 0) {
-      ctx->error_code = CSYNC_ERR_PROPAGATE;
-      return -1;
-  }
-  csync_gettime(&finish);
-
-  CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG,
-            "Correct Id took. %.2f seconds ",
-            c_secdiff(finish, start));
-  
-
   ctx->status |= CSYNC_STATUS_PROPAGATE;
 
   return 0;
@@ -566,8 +583,11 @@ static int _csync_treewalk_visitor(void *obj, void *data) {
     cur = (csync_file_stat_t *) obj;
     ctx = (CSYNC *) data;
 
+    if (ctx == NULL) {
+      return -1;
+    }
 
-    if (!(ctx && obj && data)) {
+    if (obj == NULL || data == NULL) {
       ctx->error_code = CSYNC_ERR_PARAM;
       return -1;
     }
@@ -594,9 +614,15 @@ static int _csync_treewalk_visitor(void *obj, void *data) {
       trav.type =   cur->type;
       trav.instruction = cur->instruction;
       trav.rename_path = cur->destpath;
+      trav.md5 =    cur->md5;
+      trav.error_string = cur->error_string;
 
       rc = (*visitor)(&trav, twctx->userdata);
       cur->instruction = trav.instruction;
+      if (trav.md5 != cur->md5) {
+          SAFE_FREE(cur->md5);
+          cur->md5 = c_strdup(trav.md5);
+      }
       return rc;
     }
     ctx->error_code = CSYNC_ERR_TREE;
@@ -681,24 +707,14 @@ static void _tree_destructor(void *data) {
   csync_file_stat_t *freedata = NULL;
 
   freedata = (csync_file_stat_t *) data;
-  SAFE_FREE(freedata->md5);
-  SAFE_FREE(freedata->destpath);
-  SAFE_FREE(freedata);
+  csync_file_stat_free(freedata);
 }
 
-int csync_destroy(CSYNC *ctx) {
+static int  _merge_and_write_statedb(CSYNC *ctx) {
   struct timespec start, finish;
-  char *lock = NULL;
   char errbuf[256] = {0};
   int jwritten = 0;
-
-  if (ctx == NULL) {
-    errno = EBADF;
-    return -1;
-  }
-  ctx->error_code = CSYNC_ERR_NONE;
-
-  csync_vio_shutdown(ctx);
+  int rc = 0;
 
   /* if we have a statedb */
   if (ctx->statedb.db != NULL) {
@@ -706,9 +722,10 @@ int csync_destroy(CSYNC *ctx) {
     if (ctx->status >= CSYNC_STATUS_DONE) {
       /* merge trees */
       if (csync_merge_file_trees(ctx) < 0) {
-        strerror_r(errno, errbuf, sizeof(errbuf));
+        C_STRERROR(errno, errbuf, sizeof(errbuf));
         CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR, "Unable to merge trees: %s",
                   errbuf);
+        rc = -1;
       } else {
         csync_gettime(&start);
         /* write the statedb to disk */
@@ -719,45 +736,161 @@ int csync_destroy(CSYNC *ctx) {
               "Writing the statedb of %zu files to disk took %.2f seconds",
               c_rbtree_size(ctx->local.tree), c_secdiff(finish, start));
         } else {
-          strerror_r(errno, errbuf, sizeof(errbuf));
+          C_STRERROR(errno, errbuf, sizeof(errbuf));
           CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR, "Unable to write statedb: %s",
                     errbuf);
+          rc = -1;
         }
       }
     }
-    csync_statedb_close(ctx, ctx->statedb.file, jwritten);
+    if (csync_statedb_close(ctx, ctx->statedb.file, jwritten) < 0) {
+      CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "ERR: closing of statedb failed.");
+      rc = -1;
+    }
+  }
+  return rc;
+}
+
+/* reset all the list to empty.
+ * used by csync_commit and csync_destroy */
+static void _csync_clean_ctx(CSYNC *ctx)
+{
+    c_list_t * walk;
+
+    while (ctx->progress_info) {
+        csync_progressinfo_t *next = ctx->progress_info->next;
+        csync_statedb_free_progressinfo(ctx->progress_info);
+        ctx->progress_info = next;
+    }
+
+    /* destroy the rbtrees */
+    if (c_rbtree_size(ctx->local.tree) > 0) {
+        c_rbtree_destroy(ctx->local.tree, _tree_destructor);
+    }
+
+    if (c_rbtree_size(ctx->remote.tree) > 0) {
+        c_rbtree_destroy(ctx->remote.tree, _tree_destructor);
+    }
+
+    csync_rename_destroy(ctx);
+
+    for (walk = c_list_last(ctx->local.ignored_cleanup); walk != NULL; walk = c_list_prev(walk)) {
+        SAFE_FREE(walk->data);
+    }
+    for (walk = c_list_last(ctx->remote.ignored_cleanup); walk != NULL; walk = c_list_prev(walk)) {
+        SAFE_FREE(walk->data);
+    }
+
+    /* free memory */
+    c_rbtree_free(ctx->local.tree);
+    c_list_free(ctx->local.list);
+    c_list_free(ctx->local.ignored_cleanup);
+    c_rbtree_free(ctx->remote.tree);
+    c_list_free(ctx->remote.list);
+    c_list_free(ctx->remote.ignored_cleanup);
+
+    ctx->remote.list = 0;
+    ctx->local.list = 0;
+    ctx->remote.ignored_cleanup = 0;
+    ctx->local.ignored_cleanup = 0;
+
+    SAFE_FREE(ctx->statedb.file);
+}
+
+int csync_commit(CSYNC *ctx) {
+  int rc = 0;
+  char *lock = NULL;
+
+
+  /* maybe the propagate was done using another propagator, let the merger think it has been done */
+  if (ctx->error_code == CSYNC_ERR_NONE)
+    ctx->status = CSYNC_STATUS_DONE;
+
+  ctx->error_code = CSYNC_ERR_NONE;
+
+  if (_merge_and_write_statedb(ctx) < 0) {
+    CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "Merge and Write database failed!");
+    if (ctx->error_code == CSYNC_ERR_NONE) {
+      ctx->error_code = CSYNC_ERR_STATEDB_WRITE;
+    }
+    rc = 1;  /* Set to soft error. */
+    /* The other steps happen anyway, what else can we do? */
   }
 
-  /* clear exclude list */
+  csync_vio_commit(ctx);
+
+  _csync_clean_ctx(ctx);
+
+  ctx->remote.read_from_db = 0;
+
+  /* Create new trees */
+  if (c_rbtree_create(&ctx->local.tree, _key_cmp, _data_cmp) < 0) {
+    ctx->error_code = CSYNC_ERR_TREE;
+    rc = -1;
+    goto out;
+  }
+
+  if (c_rbtree_create(&ctx->remote.tree, _key_cmp, _data_cmp) < 0) {
+    ctx->error_code = CSYNC_ERR_TREE;
+    rc = -1;
+    goto out;
+  }
+
+  /* reset the progress */
+  ctx->overall_progress.file_count = 0;
+  ctx->overall_progress.current_file_no = 0;
+  ctx->overall_progress.byte_sum = 0;
+  ctx->overall_progress.byte_current = 0;
+
+  ctx->status = CSYNC_STATUS_INIT;
+  ctx->error_code = CSYNC_ERR_NONE;
+  SAFE_FREE(ctx->error_string);
+
+  /* create lock file */
+  if (asprintf(&lock, "%s/%s", ctx->local.uri, CSYNC_LOCK_FILE) < 0) {
+    ctx->error_code = CSYNC_ERR_MEM;
+    rc = -1;
+    goto out;
+  }
+  csync_lock_remove(ctx, lock);
+  SAFE_FREE(lock);
+
+  out:
+  return rc;
+}
+
+int csync_destroy(CSYNC *ctx) {
+  char *lock = NULL;
+
+  if (ctx == NULL) {
+    errno = EBADF;
+    return -1;
+  }
+  ctx->error_code = CSYNC_ERR_NONE;
+
+  csync_vio_shutdown(ctx);
+
+  if (_merge_and_write_statedb(ctx) < 0) {
+    CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "destroy: Merge and Write database failed!");
+    if (ctx->error_code == CSYNC_ERR_NONE) {
+      ctx->error_code = CSYNC_ERR_STATEDB_WRITE;
+    }
+    /* The other steps happen anyway, what else can we do? */
+  }
+
+  /* destroy exclude list */
   csync_exclude_destroy(ctx);
 
-#ifndef _WIN32
   /* remove the lock file */
   if (asprintf(&lock, "%s/%s", ctx->options.config_dir, CSYNC_LOCK_FILE) > 0) {
     csync_lock_remove(ctx, lock);
   }
-#endif
 
-  /* destroy the rbtrees */
-  if (c_rbtree_size(ctx->local.tree) > 0) {
-    c_rbtree_destroy(ctx->local.tree, _tree_destructor);
-  }
+  _csync_clean_ctx(ctx);
 
-  if (c_rbtree_size(ctx->remote.tree) > 0) {
-    c_rbtree_destroy(ctx->remote.tree, _tree_destructor);
-  }
-
-  /* free memory */
-  c_rbtree_free(ctx->local.tree);
-  c_list_free(ctx->local.list);
-  c_list_free(ctx->local.id_list);
-  c_rbtree_free(ctx->remote.tree);
-  c_list_free(ctx->remote.list);
-  c_list_free(ctx->remote.id_list);
   SAFE_FREE(ctx->local.uri);
   SAFE_FREE(ctx->remote.uri);
   SAFE_FREE(ctx->options.config_dir);
-  SAFE_FREE(ctx->statedb.file);
   SAFE_FREE(ctx->error_string);
 
 #ifdef WITH_ICONV
@@ -784,16 +917,19 @@ int csync_add_exclude_list(CSYNC *ctx, const char *path) {
   if (ctx == NULL || path == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   return csync_exclude_load(ctx, path);
 }
 
+void csync_clear_exclude_list(CSYNC *ctx)
+{
+    csync_exclude_clear(ctx);
+}
+
 const char *csync_get_config_dir(CSYNC *ctx) {
   if (ctx == NULL) {
     return NULL;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   return ctx->options.config_dir;
 }
@@ -802,7 +938,6 @@ int csync_set_config_dir(CSYNC *ctx, const char *path) {
   if (ctx == NULL || path == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   SAFE_FREE(ctx->options.config_dir);
   ctx->options.config_dir = c_strdup(path);
@@ -821,7 +956,9 @@ int csync_enable_statedb(CSYNC *ctx) {
   ctx->error_code = CSYNC_ERR_NONE;
 
   if (ctx->status & CSYNC_STATUS_INIT) {
-    ctx->error_code = CSYNC_ERR_UNSPEC;
+      if (ctx->error_code == CSYNC_ERR_NONE) {
+          ctx->error_code = CSYNC_ERR_UNSPEC;
+      }
     fprintf(stderr, "csync_enable_statedb: This function must be called before initialization.\n");
     return -1;
   }
@@ -852,8 +989,6 @@ int csync_is_statedb_disabled(CSYNC *ctx) {
   if (ctx == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
-
   return ctx->statedb.disabled;
 }
 
@@ -861,14 +996,12 @@ int csync_set_auth_callback(CSYNC *ctx, csync_auth_callback cb) {
   if (ctx == NULL || cb == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   if (ctx->status & CSYNC_STATUS_INIT) {
     fprintf(stderr, "csync_set_auth_callback: This function must be called before initialization.\n");
     ctx->error_code = CSYNC_ERR_UNSPEC;
     return -1;
   }
-
   ctx->callbacks.auth_function = cb;
 
   return 0;
@@ -897,11 +1030,6 @@ int csync_set_log_callback(CSYNC *ctx, csync_log_callback cb) {
     return -1;
   }
 
-  if (ctx->status & CSYNC_STATUS_INIT) {
-    fprintf(stderr, "csync_set_log_callback: This function must be called before initialization.\n");
-    return -1;
-  }
-
   ctx->callbacks.log_function = cb;
 
   return 0;
@@ -920,8 +1048,6 @@ void *csync_get_userdata(CSYNC *ctx) {
   if (ctx == NULL) {
     return NULL;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
-
   return ctx->callbacks.userdata;
 }
 
@@ -929,7 +1055,6 @@ int csync_set_userdata(CSYNC *ctx, void *userdata) {
   if (ctx == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   ctx->callbacks.userdata = userdata;
 
@@ -940,7 +1065,6 @@ csync_auth_callback csync_get_auth_callback(CSYNC *ctx) {
   if (ctx == NULL) {
     return NULL;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   return ctx->callbacks.auth_function;
 }
@@ -957,7 +1081,6 @@ int csync_set_status(CSYNC *ctx, int status) {
   if (ctx == NULL || status < 0) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   ctx->status = status;
 
@@ -968,7 +1091,6 @@ int csync_get_status(CSYNC *ctx) {
   if (ctx == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   return ctx->status;
 }
@@ -977,7 +1099,6 @@ int csync_enable_conflictcopys(CSYNC* ctx){
   if (ctx == NULL) {
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
 
   if (ctx->status & CSYNC_STATUS_INIT) {
     fprintf(stderr, "csync_enable_conflictcopys: This function must be called before initialization.\n");
@@ -1051,23 +1172,62 @@ int csync_set_iconv_codec(const char *from)
 }
 #endif
 
+void csync_request_abort(CSYNC *ctx)
+{
+  if (ctx != NULL) {
+    ctx->abort = true;
+  }
+}
+
+void csync_resume(CSYNC *ctx)
+{
+  if (ctx != NULL) {
+    ctx->abort = false;
+  }
+}
+
+int  csync_abort_requested(CSYNC *ctx)
+{
+  if (ctx != NULL) {
+    return ctx->abort;
+  } else {
+    return (1 == 0);
+  }
+}
+
+void csync_file_stat_free(csync_file_stat_t *st)
+{
+  if (st) {
+    SAFE_FREE(st->md5);
+    SAFE_FREE(st->error_string);
+    SAFE_FREE(st->destpath);
+    SAFE_FREE(st);
+  }
+}
+
 int csync_set_progress_callback(CSYNC* ctx, csync_progress_callback cb)
 {
-  if (ctx == NULL || cb == NULL) {
+  if (ctx == NULL) {
+    return -1;
+  }
+  if (cb == NULL ) {
     ctx->error_code = CSYNC_ERR_PARAM;
     return -1;
   }
-  ctx->error_code = CSYNC_ERR_NONE;
-  ctx->callbacks.progresscb = cb;
 
-  if (ctx->status & CSYNC_STATUS_INIT) {
-    fprintf(stderr, "csync_set_progress_callback: This function must be called before initialization.\n");
-    ctx->error_code = CSYNC_ERR_UNSPEC;
-    return -1;
-  }
+  ctx->callbacks.progress_cb = cb;
 
   return 0;
 
 }
 
+csync_progress_callback csync_get_progress_callback(CSYNC *ctx)
+{
+  if (ctx == NULL) {
+    return NULL;
+  }
+
+  return ctx->callbacks.progress_cb;
+}
+
 /* vim: set ts=8 sw=2 et cindent: */
