@@ -39,6 +39,7 @@
 #include "c_lib.h"
 #include "c_private.h"
 #include "csync.h"
+#include "csync_misc.h"
 
 #ifdef WITH_ICONV
 #include <iconv.h>
@@ -81,7 +82,6 @@ enum csync_replica_e {
 
 typedef struct csync_file_stat_s csync_file_stat_t;
 
-
 /**
  * @brief csync public structure
  */
@@ -89,7 +89,7 @@ struct csync_s {
   struct {
       csync_auth_callback auth_function;
       csync_log_callback log_function;
-      csync_progress_callback progresscb;
+      csync_progress_callback progress_cb;
       void *userdata;
   } callbacks;
   c_strlist_t *excludes;
@@ -106,6 +106,7 @@ struct csync_s {
     c_rbtree_t *tree;
     c_list_t *list;
     enum csync_replica_e type;
+    c_list_t *ignored_cleanup;
   } local;
 
   struct {
@@ -114,6 +115,7 @@ struct csync_s {
     c_list_t *list;
     enum csync_replica_e type;
     int  read_from_db;
+    c_list_t *ignored_cleanup;
   } remote;
 
   struct {
@@ -144,7 +146,9 @@ struct csync_s {
     uid_t euid;
   } pwd;
 
-  struct csync_progressinfo_s *progress;
+  csync_overall_progress_t overall_progress;
+
+  struct csync_progressinfo_s *progress_info;
 
   /* replica we are currently walking */
   enum csync_replica_e current;
@@ -172,7 +176,7 @@ struct csync_s {
 struct csync_file_stat_s {
   uint64_t phash;   /* u64 */
   time_t modtime;   /* u64 */
-  off_t size;       /* u64 */
+  int64_t size;       /* u64 */
   size_t pathlen;   /* u64 */
   uint64_t inode;   /* u64 */
   uid_t uid;        /* u32 */
@@ -198,6 +202,8 @@ __attribute__ ((packed))
 #endif
 ;
 
+void csync_file_stat_free(csync_file_stat_t *st);
+
 /*
  * context for the treewalk function
  */
