@@ -34,7 +34,7 @@
  *
  * In the first phase we copy the file to a temporary file on the opposite
  * replica. This has the advantage that we can check if file which has been
- * copied to the opposite replica has been transfered successfully. If the
+ * copied to the opposite replica has been transferred successfully. If the
  * connection gets interrupted during the transfer we still have the original
  * states of the file. This means no data will be lost.
  *
@@ -66,6 +66,11 @@
 int csync_propagate_files(CSYNC *ctx);
 
 int csync_propagate_rename_file(CSYNC *ctx, csync_file_stat_t *st);
+
+int csync_init_progress(CSYNC *ctx);
+
+void csync_finalize_progress(CSYNC *ctx);
+
 /**
  * }@
  */
