@@ -30,6 +30,8 @@
 #define CSYNC_LOG_CATEGORY_NAME "csync.reconciler"
 #include "csync_log.h"
 
+#include "inttypes.h"
+
 #define ACCEPTED_TIME_DIFF 5
 #define ONE_HOUR 3600
 
@@ -108,7 +110,7 @@ static int _csync_merge_algorithm_visitor(void *obj, void *data) {
                         len = strlen( tmp->path );
                         h = c_jhash64((uint8_t *) tmp->path, len, 0);
 
-                        CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE,"PHash of temporar opposite: %ld", h);
+                        CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "PHash of temporary opposite: %" PRIu64, h);
                         node = c_rbtree_find(tree, &h);
                     }
                     if(node) {
