@@ -27,6 +27,7 @@
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
+#include <inttypes.h>
 
 #include "c_lib.h"
 #include "c_jhash.h"
@@ -86,6 +87,7 @@ static int _csync_detect_update(CSYNC *ctx, const char *file,
   const char *path = NULL;
   csync_file_stat_t *st = NULL;
   csync_file_stat_t *tmp = NULL;
+  CSYNC_EXCLUDE_TYPE excluded;
 
   if ((file == NULL) || (fs == NULL)) {
     errno = EINVAL;
@@ -114,6 +116,25 @@ static int _csync_detect_update(CSYNC *ctx, const char *file,
 
   len = strlen(path);
 
+  /* Check if file is excluded */
+  excluded = csync_excluded(ctx, path,type);
+  if (excluded != CSYNC_NOT_EXCLUDED) {
+    CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "%s excluded  (%d)", path, excluded);
+    if (excluded == CSYNC_FILE_EXCLUDE_AND_REMOVE) {
+      switch (ctx->current) {
+        case LOCAL_REPLICA:
+          ctx->local.ignored_cleanup = c_list_append(ctx->local.ignored_cleanup, c_strdup(path));
+          break;
+        case REMOTE_REPLICA:
+          ctx->remote.ignored_cleanup = c_list_append(ctx->remote.ignored_cleanup, c_strdup(path));
+          break;
+        default:
+          break;
+      }
+      return 0;
+    }
+  }
+
   h = _hash_of_file(ctx, file );
   if( h == 0 ) {
     return -1;
@@ -162,6 +183,10 @@ static int _csync_detect_update(CSYNC *ctx, const char *file,
     st->instruction = CSYNC_INSTRUCTION_NONE;
     goto out;
   }
+  if (excluded > CSYNC_NOT_EXCLUDED || type == CSYNC_FTW_TYPE_SLINK) {
+    st->instruction = CSYNC_INSTRUCTION_IGNORE;
+    goto out;
+  }
 
   /* Update detection: Check if a database entry exists.
    * If not, the file is either new or has been renamed. To see if it is
@@ -170,29 +195,32 @@ static int _csync_detect_update(CSYNC *ctx, const char *file,
    */
   if (csync_get_statedb_exists(ctx)) {
     tmp = csync_statedb_get_stat_by_hash(ctx, h);
-#if 0
-    /* this code could possibly replace the one in csync_vio.c stat and would be more efficient */
-    if(tmp) {
-        if( ctx->current == LOCAL_REPLICA ) {
-            if(fs->mtime == tmp->modtime && fs->size == tmp->size) {
-                /* filesystem modtime is still the same as the db mtime
-                 * thus the md5 sum is still valid. */
-                fs->md5 = c_strdup( tmp->md5 );
-            }
-        }
-    }
-#endif
+
     if(tmp && tmp->phash == h ) { /* there is an entry in the database */
         /* we have an update! */
-        CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "Database entry found, compare: %lu <-> %lu, md5: %s <-> %s",
-                  fs->mtime, tmp->modtime, fs->md5, tmp->md5);
+        CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "Database entry found, compare: %" PRId64 " <-> %" PRId64 ", md5: %s <-> %s, inode: %" PRId64 " <-> %" PRId64,
+                  ((int64_t) fs->mtime), ((int64_t) tmp->modtime), fs->md5, tmp->md5, (int64_t) fs->inode, (int64_t) tmp->inode);
         if( !fs->md5) {
             st->instruction = CSYNC_INSTRUCTION_EVAL;
             goto out;
         }
         if((ctx->current == REMOTE_REPLICA && !c_streq(fs->md5, tmp->md5 ))
-            || (ctx->current == LOCAL_REPLICA && fs->mtime != tmp->modtime)) {
-            // if (!fs->mtime > tmp->modtime) {
+            || (ctx->current == LOCAL_REPLICA && (fs->mtime != tmp->modtime
+#if 0
+                                                  || fs->inode != tmp->inode
+#endif
+                                                  ))) {
+            /* Comparison of the local inode is disabled because people reported problems
+             * on windows with flacky inode values, see github bug #779
+             *
+             * The inode needs to be observed because:
+             * $>  echo a > a.txt ; echo b > b.txt
+             * both files have the same mtime
+             * sync them.
+             * $> rm a.txt && mv b.txt a.txt
+             * makes b.txt appearing as a.txt yet a sync is not performed because
+             * both have the same modtime as mv does not change that.
+             */
             st->instruction = CSYNC_INSTRUCTION_EVAL;
             goto out;
         }
@@ -202,7 +230,7 @@ static int _csync_detect_update(CSYNC *ctx, const char *file,
         if (ctx->current == LOCAL_REPLICA) {
             tmp = csync_statedb_get_stat_by_inode(ctx, fs->inode);
             if (tmp && tmp->inode == fs->inode && (tmp->modtime == fs->mtime || fs->type == CSYNC_VIO_FILE_TYPE_DIRECTORY)) {
-                CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "inodes: %ld <-> %ld", tmp->inode, fs->inode);
+                CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "inodes: %" PRId64 " <-> %" PRId64, (int64_t) tmp->inode, (int64_t) fs->inode);
                 /* inode found so the file has been renamed */
                 st->instruction = CSYNC_INSTRUCTION_RENAME;
                 if (fs->type == CSYNC_VIO_FILE_TYPE_DIRECTORY) {
@@ -224,14 +252,21 @@ static int _csync_detect_update(CSYNC *ctx, const char *file,
 
 out:
 
+  /* Set the ignored error string. */
+  if (st->instruction == CSYNC_INSTRUCTION_IGNORE) {
+    if (excluded == CSYNC_FILE_EXCLUDE_LIST) {
+      st->error_string = c_strdup("File listed on ignore list.");
+    } else if (excluded == CSYNC_FILE_EXCLUDE_INVALID_CHAR) {
+      st->error_string = c_strdup("File contains invalid characters.");
+    }
+  }
   if (st->instruction != CSYNC_INSTRUCTION_NONE && st->instruction != CSYNC_INSTRUCTION_IGNORE
       && type != CSYNC_FTW_TYPE_DIR) {
     st->child_modified = 1;
   }
   ctx->current_fs = st;
 
-  if( tmp) SAFE_FREE(tmp->md5);
-  SAFE_FREE(tmp);
+  csync_file_stat_free(tmp);
   st->inode = fs->inode;
   st->mode  = fs->mode;
   st->size  = fs->size;
@@ -305,15 +340,15 @@ int csync_walker(CSYNC *ctx, const char *file, const csync_vio_file_stat_t *fs,
     if( !st ) {
       return 0;
     }
-    SAFE_FREE(st->md5);
-    SAFE_FREE(st->destpath);
-    SAFE_FREE(st->error_string);
-    SAFE_FREE(st);
+    csync_file_stat_free(st);
+    st = NULL;
 
     type = CSYNC_FTW_TYPE_SKIP;
     break;
   case CSYNC_FTW_FLAG_SLINK:
     CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "symlink: %s - not supported", file);
+    type = CSYNC_FTW_TYPE_SLINK;
+    break;
   case CSYNC_FTW_FLAG_DNR:
   case CSYNC_FTW_FLAG_DP:
   case CSYNC_FTW_FLAG_SLN:
@@ -426,7 +461,7 @@ int csync_ftw(CSYNC *ctx, const char *uri, csync_walker_fn fn,
       ctx->error_code = CSYNC_ERR_PROXY;
       goto error;
     } else {
-      strerror_r(errno, errbuf, sizeof(errbuf));
+      C_STRERROR(errno, errbuf, sizeof(errbuf));
       CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR,
           "opendir failed for %s - %s (errno %d)",
           uri, errbuf, errno);
@@ -460,7 +495,7 @@ int csync_ftw(CSYNC *ctx, const char *uri, csync_walker_fn fn,
       goto error;
     }
 
-    /* Create relative path for checking the exclude list */
+    /* Create relative path */
     switch (ctx->current) {
       case LOCAL_REPLICA:
         path = filename + strlen(ctx->local.uri) + 1;
@@ -472,9 +507,8 @@ int csync_ftw(CSYNC *ctx, const char *uri, csync_walker_fn fn,
         break;
     }
 
-    /* Check if file is excluded */
-    if (csync_excluded(ctx, path)) {
-      CSYNC_LOG(CSYNC_LOG_PRIORITY_TRACE, "%s excluded", path);
+    /* skip ".csync_journal.db" and ".csync_journal.db.ctmp" */
+    if (c_streq(path, ".csync_journal.db") || c_streq(path, ".csync_journal.db.ctmp")) {
       csync_vio_file_stat_destroy(dirent);
       dirent = NULL;
       SAFE_FREE(filename);
@@ -524,8 +558,11 @@ int csync_ftw(CSYNC *ctx, const char *uri, csync_walker_fn fn,
             fs->md5 = md5;
             fs->fields |= CSYNC_VIO_FILE_STAT_FIELDS_MD5;
         }
-
-        CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "Uniq ID from Database: %s -> %s", path, fs->md5 ? fs->md5 : "<NULL>" );
+        if( c_streq(md5, "")) {
+          CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "Uniq ID from Database is EMPTY: %s", path);
+        } else {
+          CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "Uniq ID from Database: %s -> %s", path, fs->md5 ? fs->md5 : "<NULL>" );
+        }
     }
 
     previous_fs = ctx->current_fs;
@@ -533,13 +570,15 @@ int csync_ftw(CSYNC *ctx, const char *uri, csync_walker_fn fn,
     /* Call walker function for each file */
     rc = fn(ctx, filename, fs, flag);
 
-    if (ctx->current_fs && previous_fs && ctx->current_fs->child_modified)
+    if (ctx->current_fs && previous_fs && ctx->current_fs->child_modified) {
         previous_fs->child_modified = ctx->current_fs->child_modified;
+    }
 
-    if( ! do_read_from_db )
+    if( ! do_read_from_db ) {
         csync_vio_file_stat_destroy(fs);
-    else
+    } else {
         SAFE_FREE(fs->md5);
+    }
 
     if (rc < 0) {
       csync_vio_closedir(ctx, dh);
