@@ -34,6 +34,7 @@
 
 #define CSYNC_LOG_CATEGORY_NAME "csync.util"
 #include "csync_log.h"
+#include "csync_statedb.h"
 
 typedef struct {
   const char *instr_str;
@@ -143,7 +144,7 @@ static int _merge_file_trees_visitor(void *obj, void *data) {
 
     new_stat = c_malloc(sizeof(csync_file_stat_t) + fs->pathlen + 1);
     if (new_stat == NULL) {
-      strerror_r(errno, errbuf, sizeof(errbuf));
+      C_STRERROR(errno, errbuf, sizeof(errbuf));
       CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR,
           "file: %s, merge malloc, error: %s",
           fs->path,
@@ -152,15 +153,18 @@ static int _merge_file_trees_visitor(void *obj, void *data) {
       goto out;
     }
     new_stat = memcpy(new_stat, fs, sizeof(csync_file_stat_t) + fs->pathlen + 1);
-    if (fs->md5)
+    if (fs->md5) {
         new_stat->md5 = c_strdup(fs->md5);
-    if (fs->destpath)
+    }
+    if (fs->destpath) {
         new_stat->destpath = c_strdup(fs->destpath);
-    if (fs->error_string)
+    }
+    if (fs->error_string) {
         new_stat->error_string = c_strdup(fs->error_string);
+    }
 
     if (c_rbtree_insert(tree, new_stat) < 0) {
-      strerror_r(errno, errbuf, sizeof(errbuf));
+      C_STRERROR(errno, errbuf, sizeof(errbuf));
       SAFE_FREE(new_stat);
       CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR,
           "file: %s, rb tree insert, error: %s",
@@ -183,12 +187,19 @@ static int _merge_file_trees_visitor(void *obj, void *data) {
   case LOCAL_REPLICA:
       /* If there is a destpath, this is a rename and the target must be used. */
       if( fs->destpath ) {
-          asprintf(&uri, "%s/%s", ctx->local.uri, fs->destpath);
+          if (asprintf(&uri, "%s/%s", ctx->local.uri, fs->destpath) < 0) {
+              rc = -1;
+              C_STRERROR(errno, errbuf, sizeof(errbuf));
+              CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR, "file uri alloc failed: %s",
+                        errbuf);
+              SAFE_FREE(fs->destpath);
+              goto out;
+          }
           SAFE_FREE(fs->destpath);
       } else {
           if (asprintf(&uri, "%s/%s", ctx->local.uri, fs->path) < 0) {
               rc = -1;
-              strerror_r(errno, errbuf, sizeof(errbuf));
+              C_STRERROR(errno, errbuf, sizeof(errbuf));
               CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR, "file uri alloc failed: %s",
                         errbuf);
               goto out;
@@ -198,7 +209,7 @@ static int _merge_file_trees_visitor(void *obj, void *data) {
     case REMOTE_REPLICA:
       if (asprintf(&uri, "%s/%s", ctx->remote.uri, fs->path) < 0) {
           rc = -1;
-          strerror_r(errno, errbuf, sizeof(errbuf));
+          C_STRERROR(errno, errbuf, sizeof(errbuf));
           CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR, "file uri alloc failed: %s",
                     errbuf);
           goto out;
@@ -211,7 +222,7 @@ static int _merge_file_trees_visitor(void *obj, void *data) {
   /* get file stat of the file on the replica */
   vst = csync_vio_file_stat_new();
   if (csync_vio_stat(ctx, uri, vst) < 0) {
-    strerror_r(errno, errbuf, sizeof(errbuf));
+    C_STRERROR(errno, errbuf, sizeof(errbuf));
     CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR,
         "file: %s, updating stat failed, error: %s",
         uri,
@@ -243,7 +254,9 @@ static int _merge_file_trees_visitor(void *obj, void *data) {
 
   CSYNC_LOG(CSYNC_LOG_PRIORITY_DEBUG, "file: %s, instruction: UPDATED (%s)", uri, fs->md5);
 
-  fs->instruction = CSYNC_INSTRUCTION_NONE;
+  if (fs->instruction != CSYNC_INSTRUCTION_ERROR) {
+      fs->instruction = CSYNC_INSTRUCTION_NONE;
+  }
 
   rc = 0;
 out:
@@ -257,6 +270,49 @@ out:
   return rc;
 }
 
+static int _fix_errors_visitor(void *obj, void *data) {
+  csync_file_stat_t *fs = NULL;
+  csync_file_stat_t *tfs = NULL;
+
+  CSYNC *ctx = NULL;
+
+  int rc = -1;
+
+  fs = (csync_file_stat_t *) obj;
+  ctx = (CSYNC *) data;
+
+  /* search for ERROR */
+  if (fs->instruction != CSYNC_INSTRUCTION_ERROR
+      || fs->type == CSYNC_FTW_TYPE_DIR) {
+    rc = 0;
+    goto out;
+  }
+
+  /* check if the file is new or has been synced */
+  tfs = csync_statedb_get_stat_by_hash(ctx, fs->phash);
+  if (tfs == NULL) {
+    rc = 0;
+    goto out;
+  }
+
+  /* update file stat */
+  fs->inode = tfs->inode;
+  fs->modtime = tfs->modtime;
+
+  fs->instruction = CSYNC_INSTRUCTION_UPDATED;
+
+  rc = 0;
+out:
+  csync_file_stat_free(tfs);
+
+  if (rc != 0) {
+    fs->instruction = CSYNC_INSTRUCTION_ERROR;
+  }
+
+  return rc;
+}
+
+
 /*
  * merge the local tree with the new files from remote and update the
  * inode numbers
@@ -264,6 +320,15 @@ out:
 int csync_merge_file_trees(CSYNC *ctx) {
   int rc = -1;
 
+  /* walk over the local tree, in case of error take the value from the database  */
+  ctx->current = REMOTE_REPLICA;
+  ctx->replica = ctx->remote.type;
+
+  rc = c_rbtree_walk(ctx->local.tree, ctx, _fix_errors_visitor);
+  if (rc < 0) {
+    goto out;
+  }
+
   /* walk over remote tree, stat on local system */
   ctx->current = LOCAL_REPLICA;
   ctx->replica = ctx->local.type;
@@ -272,20 +337,6 @@ int csync_merge_file_trees(CSYNC *ctx) {
   if (rc < 0) {
     goto out;
   }
-
-#if 0
-  /* We don't have to merge the remote tree atm. */
-
-  /* walk over local tree, stat on remote system */
-  ctx->current = REMOTE_REPLICA;
-  ctx->replica = ctx->remote.type;
-
-  rc = c_rbtree_walk(ctx->local.tree, ctx, _merge_file_trees_visitor);
-  if (rc < 0) {
-    goto out;
-  }
-#endif
-
 out:
   return rc;
 }
@@ -328,11 +379,11 @@ out:
 void csync_win32_set_file_hidden( const char *file, bool h ) {
 #ifdef _WIN32
   const _TCHAR *fileName;
+  DWORD dwAttrs;
   if( !file ) return;
 
   fileName = c_multibyte( file );
-
-  DWORD dwAttrs = GetFileAttributesW(fileName);
+  dwAttrs = GetFileAttributesW(fileName);
 
   if (dwAttrs==INVALID_FILE_ATTRIBUTES) return;
 
