@@ -22,11 +22,13 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <sys/timeb.h>
 #include <sys/time.h>
+#include <inttypes.h>
 
 #include "httpbf.h"
 
@@ -34,15 +36,29 @@
 #include <neon/ne_request.h>
 #include <neon/ne_basic.h>
 
-#ifdef NDEBUG
-#define DEBUG_HBF(...)
-#else
-#define DEBUG_HBF(...) printf(__VA_ARGS__)
-#endif
+// #ifdef NDEBUG
+// #define DEBUG_HBF(...)
+// #else
+#define DEBUG_HBF(...) { if(transfer->log_cb) { \
+        char buf[1024];                         \
+        snprintf(buf, 1024, __VA_ARGS__);       \
+        transfer->log_cb(__FUNCTION__, buf);    \
+  }  }
+
+// #endif
 
 #define DEFAULT_BLOCK_SIZE (10*1024*1024)
 
-static int transfer_id( struct stat *sb ) {
+/* Platform specific defines go here. */
+#ifdef _WIN32
+#define _hbf_fstat _fstat64
+typedef struct stat64 hbf_stat_t;
+#else
+#define _hbf_fstat fstat
+typedef struct stat hbf_stat_t;
+#endif
+
+static int transfer_id( hbf_stat_t *sb ) {
     struct timeval tp;
     int res;
     int r;
@@ -78,16 +94,17 @@ hbf_transfer_t *hbf_init_transfer( const char *dest_uri ) {
     transfer->error_string = NULL;
     transfer->start_id = 0;
     transfer->block_size = DEFAULT_BLOCK_SIZE;
+    transfer->threshold = transfer->block_size;
 
     return transfer;
 }
 
 /* Create the splitlist of a given file descriptor */
 Hbf_State hbf_splitlist(hbf_transfer_t *transfer, int fd ) {
-  struct stat sb;
-  off_t num_blocks;
-  off_t blk_size;
-  off_t remainder = 0;
+  hbf_stat_t sb;
+  int64_t num_blocks;
+  int64_t blk_size;
+  int64_t remainder = 0;
 
   if( ! transfer ) {
       return HBF_PARAM_FAIL;
@@ -98,7 +115,7 @@ Hbf_State hbf_splitlist(hbf_transfer_t *transfer, int fd ) {
     return HBF_PARAM_FAIL;
   }
   
-  if( fstat(fd, &sb) < 0 ) {
+  if( _hbf_fstat(fd, &sb) < 0 ) {
     DEBUG_HBF("Failed to stat the file descriptor: errno = %d", errno);
     return HBF_FILESTAT_FAIL;
   }
@@ -111,8 +128,15 @@ Hbf_State hbf_splitlist(hbf_transfer_t *transfer, int fd ) {
   transfer->calc_size = 0;
 #endif
 
+  DEBUG_HBF("hbf_splitlist: block_size: %" PRId64 " threshold: %" PRId64 " st_size: %" PRId64 "\n", transfer->block_size, transfer->threshold, sb.st_size );
+
+
   /* calc the number of blocks to split in */
   blk_size = transfer->block_size;
+  if (sb.st_size < transfer->threshold) {
+      blk_size = transfer->threshold;
+  }
+
   num_blocks = sb.st_size / blk_size;
 
   /* there migth be a remainder. */
@@ -129,9 +153,12 @@ Hbf_State hbf_splitlist(hbf_transfer_t *transfer, int fd ) {
     blk_size   = 0;
   }
 
+  DEBUG_HBF("hbf_splitlist: num_blocks: %" PRId64 " rmainder: %" PRId64 " blk_size: %" PRId64 "\n", num_blocks, remainder, blk_size );
+
+
   if( num_blocks ) {
       int cnt;
-      off_t overall = 0;
+      int64_t overall = 0;
       /* create a datastructure for the transfer data */
       transfer->block_arr = calloc(num_blocks, sizeof(hbf_block_t*));
       transfer->block_cnt = num_blocks;
@@ -157,7 +184,10 @@ Hbf_State hbf_splitlist(hbf_transfer_t *transfer, int fd ) {
           overall += block->size;
           /* store the block data into the result array in the transfer */
           *((transfer->block_arr)+cnt) = block;
+
+          DEBUG_HBF("hbf_splitlist: created block %d   (start: %" PRId64 "  size: %" PRId64 ")", cnt, block->start, block->size );
       }
+
 #ifndef NDEBUG
   transfer->calc_size = overall;
 #endif
@@ -245,16 +275,18 @@ static char* get_transfer_url( hbf_transfer_t *transfer, int indx ) {
  * returns HBF_TRANSFER_SUCCESS if the transfer of this block was a success
  * returns HBF_SUCCESS if the server aknoweldge that he received all the blocks
  */
-static int dav_request( ne_request *req, int fd, hbf_block_t *blk ) {
+static int _hbf_dav_request(hbf_transfer_t *transfer, ne_request *req, int fd, hbf_block_t *blk ) {
     Hbf_State state = HBF_TRANSFER_SUCCESS;
     int res;
     const ne_status *req_status = NULL;
     const char *etag = NULL;
 
+    (void) transfer;
+
     if( ! (blk && req) ) return HBF_PARAM_FAIL;
 
     ne_set_request_body_fd(req, fd, blk->start, blk->size);
-    /* DEBUG_HBF("HBF: Block: %d , Start: %ld and Size: %ld\n", blk->seq_number, blk->start, blk->size ); */
+    DEBUG_HBF("HBF: Block: %d , Start: %" PRId64 " and Size: %" PRId64 "", blk->seq_number, blk->start, blk->size );
     res = ne_request_dispatch(req);
 
     req_status = ne_get_status( req );
@@ -280,6 +312,8 @@ static int dav_request( ne_request *req, int fd, hbf_block_t *blk ) {
                 } else {
                     blk->etag = strdup( etag );
                 }
+            } else {
+                /* DEBUG_HBF("OOOOOOOO No etag returned!"); */
             }
         }
         break;
@@ -315,7 +349,7 @@ static int dav_request( ne_request *req, int fd, hbf_block_t *blk ) {
 
 static Hbf_State validate_source_file( hbf_transfer_t *transfer ) {
   Hbf_State state = HBF_SUCCESS;
-  struct stat sb;
+  hbf_stat_t sb;
 
   if( transfer == NULL ) {
     state = HBF_PARAM_FAIL;
@@ -328,7 +362,7 @@ static Hbf_State validate_source_file( hbf_transfer_t *transfer ) {
   }
 
   if( state == HBF_SUCCESS ) {
-    int rc = fstat( transfer->fd, &sb );
+    int rc = _hbf_fstat( transfer->fd, &sb );
     if( rc != 0 ) {
       state = HBF_STAT_FAIL;
     }
@@ -405,9 +439,10 @@ Hbf_State hbf_transfer( ne_session *session, hbf_transfer_t *transfer, const cha
         if( ! block ) state = HBF_PARAM_FAIL;
 
         if( transfer->abort_cb ) {
-            int abort = (transfer->abort_cb)();
-            if( abort ) {
+            int do_abort = (transfer->abort_cb)();
+            if( do_abort ) {
               state = HBF_USER_ABORTED;
+              transfer->start_id = block_id  % transfer->block_cnt;
             }
         }
 
@@ -432,10 +467,13 @@ Hbf_State hbf_transfer( ne_session *session, hbf_transfer_t *transfer, const cha
             ne_request *req = ne_request_create(session, verb, transfer_url);
 
             if( req ) {
+                char buf[21];
+                snprintf(buf, sizeof(buf), "%"PRId64, transfer->stat_size);
+                ne_add_request_header(req, "OC-Total-Length", buf);
                 if( transfer->block_cnt > 1 ) {
                   ne_add_request_header(req, "OC-Chunked", "1");
                 }
-                state = dav_request( req, transfer->fd, block );
+                state = _hbf_dav_request(transfer,  req, transfer->fd, block );
 
                 if( state != HBF_TRANSFER_SUCCESS && state != HBF_SUCCESS) {
                   if( transfer->error_string ) free( transfer->error_string );
@@ -469,6 +507,10 @@ Hbf_State hbf_transfer( ne_session *session, hbf_transfer_t *transfer, const cha
     /* do the source file validation finally (again). */
     if( state == HBF_SUCCESS ) {
         state = validate_source_file(transfer);
+    } else if( state == HBF_TRANSFER_SUCCESS ) {
+        /* This means that no etag was returned on one of the chunks to indicate
+         * that the upload was finished. */
+        state = HBF_TRANSFER_NOT_ACKED;
     }
 
     return state;
@@ -491,9 +533,10 @@ int hbf_fail_http_code( hbf_transfer_t *transfer )
   return 200;
 }
 
-const char *hbf_error_string( Hbf_State state )
+const char *hbf_error_string(hbf_transfer_t *transfer, Hbf_State state)
 {
     const char *re;
+    int cnt;
     switch( state ) {
     case HBF_SUCCESS:
         re = "Ok.";
@@ -502,7 +545,7 @@ const char *hbf_error_string( Hbf_State state )
         re = "Block was not yet tried to transfer.";
         break;
     case HBF_TRANSFER:         /* transfer currently running  */
-        re = "Block is currently transfered.";
+        re = "Block is currently transferred.";
         break;
     case HBF_TRANSFER_FAILED:  /* transfer tried but failed   */
         re = "Block transfer failed.";
@@ -537,12 +580,29 @@ const char *hbf_error_string( Hbf_State state )
     case HBF_MEMORY_FAIL:
         re = "Out of memory.";
         break;
+    case HBF_STAT_FAIL:
+        re = "Filesystem stat on file failed.";
+        break;
     case HBF_SOURCE_FILE_CHANGE:
       re = "Source file changed too often during upload.";
       break;
-
+    case HBF_USER_ABORTED:
+        re = "Transmission aborted by user.";
+        break;
+    case HBF_TRANSFER_NOT_ACKED:
+        re = "The server did not provide an Etag.";
+        break;
     case HBF_FAIL:
     default:
+        for( cnt = 0; cnt < transfer->block_cnt; cnt++ ) {
+            int block_id = (cnt + transfer->start_id) % transfer->block_cnt;
+            hbf_block_t *block = transfer->block_arr[block_id];
+
+            if( block->state != HBF_NOT_TRANSFERED && block->state != HBF_TRANSFER_SUCCESS
+                    && block->http_error_msg != NULL) {
+                return block->http_error_msg;
+            }
+        }
         re = "Unknown error.";
     }
     return re;
@@ -554,3 +614,10 @@ void hbf_set_abort_callback( hbf_transfer_t *transfer, hbf_abort_callback cb)
     transfer->abort_cb = cb;
   }
 }
+
+void hbf_set_log_callback(hbf_transfer_t* transfer, hbf_log_callback cb)
+{
+    if( transfer ) {
+        transfer->log_cb = cb;
+    }
+}
