@@ -31,6 +31,7 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <string.h>
+#include <stdio.h>
 
 #include "c_file.h"
 #include "c_string.h"
@@ -60,6 +61,27 @@ int c_isfile(const char *path) {
 }
 
 /* copy file from src to dst, overwrites dst */
+#ifdef _WIN32
+int c_copy(const char* src, const char *dst, mode_t mode) {
+  int rc = -1;
+  _TCHAR *wsrc = 0;
+  _TCHAR *wdst = 0;
+  (void) mode; /* unused on win32 */
+  if(src && dst) {
+    wsrc = c_multibyte(src);
+    wdst = c_multibyte(dst);
+    if (CopyFileW(wsrc, wdst, FALSE)) {
+      rc = 0;
+    }
+    c_free_multibyte(wsrc);
+    c_free_multibyte(wdst);
+    if( rc < 0 ) {
+      errno = GetLastError();
+    }
+  }
+  return rc;
+}
+#else
 int c_copy(const char* src, const char *dst, mode_t mode) {
   int srcfd = -1;
   int dstfd = -1;
@@ -68,23 +90,6 @@ int c_copy(const char* src, const char *dst, mode_t mode) {
   csync_stat_t sb;
   char buf[4096];
 
-#ifdef _WIN32
-  if(src && dst) {
-      _TCHAR *wsrc = c_multibyte(src);
-      _TCHAR *wdst = c_multibyte(dst);
-      if (CopyFileW(wsrc, wdst, FALSE)) {
-          rc = 0;
-      }
-      c_free_multibyte(wsrc);
-      c_free_multibyte(wdst);
-
-      if( rc < 0 ) {
-          errno = GetLastError();
-      }
-  }
-  return rc;
-#else
-
   /* Win32 does not come here. */
   if (c_streq(src, dst)) {
     return -1;
@@ -151,17 +156,74 @@ int c_copy(const char* src, const char *dst, mode_t mode) {
 
   rc = 0;
 out:
-  if (srcfd > 0) {
+  if (srcfd >= 0) {
     close(srcfd);
   }
-  if (dstfd > 0) {
+  if (dstfd >= 0) {
       close(dstfd);
   }
   if (rc < 0) {
     unlink(dst);
   }
   return rc;
+}
+#endif // _WIN32
+
+int c_rename( const char *src, const char *dst ) {
+    _TCHAR *nuri = NULL;
+    _TCHAR *ouri = NULL;
+    int rc = 0;
+
+    nuri = c_multibyte(dst);
+    if (nuri == NULL) {
+        return -1;
+    }
+
+    ouri = c_multibyte(src);
+    if (ouri == NULL) {
+	c_free_multibyte(nuri);
+        return -1;
+    }
+
+#ifdef _WIN32
+    {
+#define MAX_TRIES_RENAME 3
+        int err = 0;
+        int cnt = 0;
+
+        do {
+            BOOL ok;
+            ok = MoveFileExW(ouri,
+                             nuri,
+                             MOVEFILE_COPY_ALLOWED +
+                             MOVEFILE_REPLACE_EXISTING +
+                             MOVEFILE_WRITE_THROUGH);
+            if (!ok) {
+                /* error */
+                err = GetLastError();
+                if( (err == ERROR_ACCESS_DENIED   ||
+                     err == ERROR_LOCK_VIOLATION  ||
+                     err == ERROR_SHARING_VIOLATION) && cnt < MAX_TRIES_RENAME ) {
+                    cnt++;
+                    Sleep(cnt*100);
+                    continue;
+                }
+            }
+            break;
+        } while( 1 );
+        if( err != 0 ) {
+            errno = err;
+            rc = -1;
+        }
+    }
+#else
+    rc = rename(ouri, nuri);
 #endif
+
+    c_free_multibyte(nuri);
+    c_free_multibyte(ouri);
+
+    return rc;
 }
 
 int c_compare_file( const char *f1, const char *f2 ) {
