@@ -106,14 +106,14 @@ int csync_vio_init(CSYNC *ctx, const char *module, const char *args) {
 
 #ifdef __APPLE__
   if (lstat(path, &sb) < 0) {
+    char path_tmp[1024];
+    char* path2 = NULL;
+    uint32_t size = sizeof(path_tmp);
     SAFE_FREE(path);
 
-    char path_tmp[1024];
-    uint32_t size = sizeof(path_tmp);
     if (_NSGetExecutablePath(path_tmp, &size) == 0)
         printf("executable path is %s\n", path_tmp);
 
-    char* path2 = NULL;
     path2 = c_dirname(path_tmp);
 
     if (asprintf(&path, "%s/../Plugins/ocsync_%s.%s", path2, module, MODULE_EXTENSION) < 0) {
@@ -366,8 +366,8 @@ int csync_vio_sendfile(CSYNC *ctx, csync_vio_handle_t *sfp, csync_vio_handle_t *
     return rc;
 }
 
-off_t csync_vio_lseek(CSYNC *ctx, csync_vio_handle_t *fhandle, off_t offset, int whence) {
-  off_t ro = 0;
+int64_t csync_vio_lseek(CSYNC *ctx, csync_vio_handle_t *fhandle, int64_t offset, int whence) {
+  int64_t ro = 0;
 
   switch(ctx->replica) {
     case REMOTE_REPLICA:
@@ -511,7 +511,7 @@ int csync_vio_mkdirs(CSYNC *ctx, const char *uri, mode_t mode) {
         return -1;
       }
     } else if (errno != ENOENT) {
-      strerror_r(errno, errbuf, sizeof(errbuf));
+      C_STRERROR(errno, errbuf, sizeof(errbuf));
       CSYNC_LOG(CSYNC_LOG_PRIORITY_ERROR, "csync_vio_mkdirs stat failed: %s",
           errbuf);
       csync_vio_file_stat_destroy(st);
